@@ -19,14 +19,15 @@
 #include "gl-shaderparser.h"
 
 static void gl_write_function_contents(struct gl_shader_parser *glsp,
-		struct cf_token **p_token, const char *end);
+				       struct cf_token **p_token,
+				       const char *end);
 
 static inline struct shader_var *sp_getparam(struct gl_shader_parser *glsp,
-		struct cf_token *token)
+					     struct cf_token *token)
 {
 	size_t i;
 	for (i = 0; i < glsp->parser.params.num; i++) {
-		struct shader_var *param = glsp->parser.params.array+i;
+		struct shader_var *param = glsp->parser.params.array + i;
 		if (strref_cmp(&token->str, param->name) == 0)
 			return param;
 	}
@@ -35,11 +36,12 @@ static inline struct shader_var *sp_getparam(struct gl_shader_parser *glsp,
 }
 
 static inline size_t sp_getsampler(struct gl_shader_parser *glsp,
-		struct cf_token *token)
+				   struct cf_token *token)
 {
 	size_t i;
 	for (i = 0; i < glsp->parser.samplers.num; i++) {
-		struct shader_sampler *sampler = glsp->parser.samplers.array+i;
+		struct shader_sampler *sampler =
+			glsp->parser.samplers.array + i;
 		if (strref_cmp(&token->str, sampler->name) == 0)
 			return i;
 	}
@@ -48,14 +50,14 @@ static inline size_t sp_getsampler(struct gl_shader_parser *glsp,
 }
 
 static inline int cmp_type(const char *name, const size_t name_len,
-		const char *type, const size_t type_len)
+			   const char *type, const size_t type_len)
 {
 	size_t min_len = (name_len < type_len) ? type_len : name_len;
 	return astrcmp_n(name, type, min_len);
 }
 
-static bool gl_write_type_n(struct gl_shader_parser *glsp,
-		const char *type, size_t len)
+static bool gl_write_type_n(struct gl_shader_parser *glsp, const char *type,
+			    size_t len)
 {
 	if (cmp_type(type, len, "float2", 6) == 0)
 		dstr_cat(&glsp->gl_string, "vec2");
@@ -90,14 +92,14 @@ static bool gl_write_type_n(struct gl_shader_parser *glsp,
 }
 
 static inline void gl_write_type(struct gl_shader_parser *glsp,
-		const char *type)
+				 const char *type)
 {
 	if (!gl_write_type_n(glsp, type, strlen(type)))
 		dstr_cat(&glsp->gl_string, type);
 }
 
 static inline bool gl_write_type_token(struct gl_shader_parser *glsp,
-		struct cf_token *token)
+				       struct cf_token *token)
 {
 	return gl_write_type_n(glsp, token->str.array, token->str.len);
 }
@@ -122,7 +124,7 @@ static inline void gl_write_params(struct gl_shader_parser *glsp)
 {
 	size_t i;
 	for (i = 0; i < glsp->parser.params.num; i++) {
-		struct shader_var *var = glsp->parser.params.array+i;
+		struct shader_var *var = glsp->parser.params.array + i;
 		gl_write_var(glsp, var);
 		dstr_cat(&glsp->gl_string, ";\n");
 	}
@@ -131,12 +133,13 @@ static inline void gl_write_params(struct gl_shader_parser *glsp)
 }
 
 static void gl_write_storage_var(struct gl_shader_parser *glsp,
-		struct shader_var *var, bool input, const char *prefix);
+				 struct shader_var *var, bool input,
+				 const char *prefix);
 
 /* unwraps a structure that's used for input/output */
 static void gl_unwrap_storage_struct(struct gl_shader_parser *glsp,
-		struct shader_struct *st, const char *name, bool input,
-		const char *prefix)
+				     struct shader_struct *st, const char *name,
+				     bool input, const char *prefix)
 {
 	struct dstr prefix_str;
 	size_t i;
@@ -148,7 +151,7 @@ static void gl_unwrap_storage_struct(struct gl_shader_parser *glsp,
 	dstr_cat(&prefix_str, "_");
 
 	for (i = 0; i < st->vars.num; i++) {
-		struct shader_var *st_var = st->vars.array+i;
+		struct shader_var *st_var = st->vars.array + i;
 		gl_write_storage_var(glsp, st_var, input, prefix_str.array);
 	}
 
@@ -156,14 +159,24 @@ static void gl_unwrap_storage_struct(struct gl_shader_parser *glsp,
 }
 
 static void gl_write_storage_var(struct gl_shader_parser *glsp,
-		struct shader_var *var, bool input, const char *prefix)
+				 struct shader_var *var, bool input,
+				 const char *prefix)
 {
-	struct shader_struct *st = shader_parser_getstruct(&glsp->parser,
-			var->type);
+	struct shader_struct *st =
+		shader_parser_getstruct(&glsp->parser, var->type);
 
 	if (st) {
 		gl_unwrap_storage_struct(glsp, st, var->name, input, prefix);
 	} else {
+		if (input && (strcmp(var->mapping, "VERTEXID") == 0))
+			return;
+		if (strcmp(var->mapping, "POSITION") == 0) {
+			if (!input && (glsp->type == GS_SHADER_VERTEX))
+				return;
+			if (input && (glsp->type == GS_SHADER_PIXEL))
+				return;
+		}
+
 		struct gl_parser_attrib attrib;
 		gl_parser_attrib_init(&attrib);
 
@@ -178,24 +191,24 @@ static void gl_write_storage_var(struct gl_shader_parser *glsp,
 		dstr_cat_dstr(&glsp->gl_string, &attrib.name);
 		dstr_cat(&glsp->gl_string, ";\n");
 
-		attrib.input   = input;
+		attrib.input = input;
 		attrib.mapping = var->mapping;
 		da_push_back(glsp->attribs, &attrib);
 	}
 }
 
 static inline void gl_write_inputs(struct gl_shader_parser *glsp,
-		struct shader_func *main)
+				   struct shader_func *main)
 {
 	size_t i;
 	for (i = 0; i < main->params.num; i++)
-		gl_write_storage_var(glsp, main->params.array+i, true,
-				"inputval_");
+		gl_write_storage_var(glsp, main->params.array + i, true,
+				     "inputval_");
 	dstr_cat(&glsp->gl_string, "\n");
 }
 
 static void gl_write_outputs(struct gl_shader_parser *glsp,
-		struct shader_func *main)
+			     struct shader_func *main)
 {
 	struct shader_var var = {0};
 	var.type = main->return_type;
@@ -208,7 +221,7 @@ static void gl_write_outputs(struct gl_shader_parser *glsp,
 }
 
 static void gl_write_struct(struct gl_shader_parser *glsp,
-		struct shader_struct *st)
+			    struct shader_struct *st)
 {
 	size_t i;
 	dstr_cat(&glsp->gl_string, "struct ");
@@ -216,7 +229,7 @@ static void gl_write_struct(struct gl_shader_parser *glsp,
 	dstr_cat(&glsp->gl_string, " {\n");
 
 	for (i = 0; i < st->vars.num; i++) {
-		struct shader_var *var = st->vars.array+i;
+		struct shader_var *var = st->vars.array + i;
 
 		dstr_cat(&glsp->gl_string, "\t");
 		gl_write_var(glsp, var);
@@ -230,7 +243,7 @@ static void gl_write_interface_block(struct gl_shader_parser *glsp)
 {
 	if (glsp->type == GS_SHADER_VERTEX) {
 		dstr_cat(&glsp->gl_string, "out gl_PerVertex {\n"
-		                           "\tvec4 gl_Position;\n};\n\n");
+					   "\tvec4 gl_Position;\n};\n\n");
 	}
 }
 
@@ -238,7 +251,7 @@ static inline void gl_write_structs(struct gl_shader_parser *glsp)
 {
 	size_t i;
 	for (i = 0; i < glsp->parser.structs.num; i++) {
-		struct shader_struct *st = glsp->parser.structs.array+i;
+		struct shader_struct *st = glsp->parser.structs.array + i;
 		gl_write_struct(glsp, st);
 	}
 }
@@ -257,6 +270,7 @@ static inline void gl_write_structs(struct gl_shader_parser *glsp)
  *   mul      -> (change to operator)
  *   rsqrt    -> inversesqrt
  *   saturate -> (use clamp)
+ *   sincos   -> (map to manual sin/cos calls)
  *   tex*     -> texture
  *   tex*grad -> textureGrad
  *   tex*lod  -> textureLod
@@ -267,13 +281,15 @@ static inline void gl_write_structs(struct gl_shader_parser *glsp)
  */
 
 static bool gl_write_mul(struct gl_shader_parser *glsp,
-		struct cf_token **p_token)
+			 struct cf_token **p_token)
 {
 	struct cf_parser *cfp = &glsp->parser.cfp;
 	cfp->cur_token = *p_token;
 
-	if (!cf_next_token(cfp))    return false;
-	if (!cf_token_is(cfp, "(")) return false;
+	if (!cf_next_token(cfp))
+		return false;
+	if (!cf_token_is(cfp, "("))
+		return false;
 
 	dstr_cat(&glsp->gl_string, "(");
 	gl_write_function_contents(glsp, &cfp->cur_token, ",");
@@ -286,14 +302,61 @@ static bool gl_write_mul(struct gl_shader_parser *glsp,
 	return true;
 }
 
+static bool gl_write_sincos(struct gl_shader_parser *glsp,
+			    struct cf_token **p_token)
+{
+	struct cf_parser *cfp = &glsp->parser.cfp;
+	struct dstr var = {0};
+	bool success = false;
+
+	cfp->cur_token = *p_token;
+
+	if (!cf_next_token(cfp))
+		return false;
+	if (!cf_token_is(cfp, "("))
+		return false;
+
+	dstr_printf(&var, "sincos_var_internal_%d", glsp->sincos_counter++);
+
+	dstr_cat(&glsp->gl_string, "float ");
+	dstr_cat_dstr(&glsp->gl_string, &var);
+	dstr_cat(&glsp->gl_string, " = ");
+	gl_write_function_contents(glsp, &cfp->cur_token, ",");
+	dstr_cat(&glsp->gl_string, "); ");
+
+	if (!cf_next_token(cfp))
+		goto fail;
+	gl_write_function_contents(glsp, &cfp->cur_token, ",");
+	dstr_cat(&glsp->gl_string, " = sin(");
+	dstr_cat_dstr(&glsp->gl_string, &var);
+	dstr_cat(&glsp->gl_string, "); ");
+
+	if (!cf_next_token(cfp))
+		goto fail;
+	gl_write_function_contents(glsp, &cfp->cur_token, ")");
+	dstr_cat(&glsp->gl_string, " = cos(");
+	dstr_cat_dstr(&glsp->gl_string, &var);
+	dstr_cat(&glsp->gl_string, ")");
+
+	success = true;
+
+fail:
+	dstr_free(&var);
+
+	*p_token = cfp->cur_token;
+	return success;
+}
+
 static bool gl_write_saturate(struct gl_shader_parser *glsp,
-		struct cf_token **p_token)
+			      struct cf_token **p_token)
 {
 	struct cf_parser *cfp = &glsp->parser.cfp;
 	cfp->cur_token = *p_token;
 
-	if (!cf_next_token(cfp))    return false;
-	if (!cf_token_is(cfp, "(")) return false;
+	if (!cf_next_token(cfp))
+		return false;
+	if (!cf_token_is(cfp, "("))
+		return false;
 
 	dstr_cat(&glsp->gl_string, "clamp");
 	gl_write_function_contents(glsp, &cfp->cur_token, ")");
@@ -304,23 +367,29 @@ static bool gl_write_saturate(struct gl_shader_parser *glsp,
 }
 
 static inline bool gl_write_texture_call(struct gl_shader_parser *glsp,
-		struct shader_var *var, const char *call, bool sampler)
+					 struct shader_var *var,
+					 const char *call, bool sampler)
 {
 	struct cf_parser *cfp = &glsp->parser.cfp;
-	size_t sampler_id = (size_t)-1;
 
-	if (!cf_next_token(cfp))    return false;
-	if (!cf_token_is(cfp, "(")) return false;
+	if (!cf_next_token(cfp))
+		return false;
+	if (!cf_token_is(cfp, "("))
+		return false;
 
 	if (sampler) {
-		if (!cf_next_token(cfp))    return false;
-		sampler_id = sp_getsampler(glsp, cfp->cur_token);
-		if (sampler_id == (size_t) -1) return false;
-		if (!cf_next_token(cfp))    return false;
-		if (!cf_token_is(cfp, ",")) return false;
-	}
+		if (!cf_next_token(cfp))
+			return false;
+		const size_t sampler_id = sp_getsampler(glsp, cfp->cur_token);
+		if (sampler_id == (size_t)-1)
+			return false;
+		if (!cf_next_token(cfp))
+			return false;
+		if (!cf_token_is(cfp, ","))
+			return false;
 
-	var->gl_sampler_id = sampler_id;
+		var->gl_sampler_id = sampler_id;
+	}
 
 	dstr_cat(&glsp->gl_string, call);
 	dstr_cat(&glsp->gl_string, "(");
@@ -331,15 +400,19 @@ static inline bool gl_write_texture_call(struct gl_shader_parser *glsp,
 
 /* processes texture.Sample(sampler, texcoord) */
 static bool gl_write_texture_code(struct gl_shader_parser *glsp,
-		struct cf_token **p_token, struct shader_var *var)
+				  struct cf_token **p_token,
+				  struct shader_var *var)
 {
 	struct cf_parser *cfp = &glsp->parser.cfp;
 	bool written = false;
 	cfp->cur_token = *p_token;
 
-	if (!cf_next_token(cfp))    return false;
-	if (!cf_token_is(cfp, ".")) return false;
-	if (!cf_next_token(cfp))    return false;
+	if (!cf_next_token(cfp))
+		return false;
+	if (!cf_token_is(cfp, "."))
+		return false;
+	if (!cf_next_token(cfp))
+		return false;
 
 	const char *function_end = ")";
 
@@ -354,13 +427,16 @@ static bool gl_write_texture_code(struct gl_shader_parser *glsp,
 	else if (cf_token_is(cfp, "Load")) {
 		written = gl_write_texture_call(glsp, var, "texelFetch", false);
 		dstr_cat(&glsp->gl_string, "(");
-		function_end = ").xy, 0)";
+		function_end = (strcmp(var->type, "texture3d") == 0)
+				       ? ").xyz, 0)"
+				       : ").xy, 0)";
 	}
 
 	if (!written)
 		return false;
 
-	if (!cf_next_token(cfp)) return false;
+	if (!cf_next_token(cfp))
+		return false;
 
 	gl_write_function_contents(glsp, &cfp->cur_token, ")");
 	dstr_cat(&glsp->gl_string, function_end);
@@ -370,13 +446,13 @@ static bool gl_write_texture_code(struct gl_shader_parser *glsp,
 }
 
 static bool gl_write_intrinsic(struct gl_shader_parser *glsp,
-		struct cf_token **p_token)
+			       struct cf_token **p_token)
 {
 	struct cf_token *token = *p_token;
 	bool written = true;
 
 	if (strref_cmp(&token->str, "atan2") == 0) {
-		dstr_cat(&glsp->gl_string, "atan2");
+		dstr_cat(&glsp->gl_string, "atan");
 	} else if (strref_cmp(&token->str, "ddx") == 0) {
 		dstr_cat(&glsp->gl_string, "dFdx");
 	} else if (strref_cmp(&token->str, "ddy") == 0) {
@@ -393,6 +469,8 @@ static bool gl_write_intrinsic(struct gl_shader_parser *glsp,
 		written = gl_write_saturate(glsp, &token);
 	} else if (strref_cmp(&token->str, "mul") == 0) {
 		written = gl_write_mul(glsp, &token);
+	} else if (strref_cmp(&token->str, "sincos") == 0) {
+		written = gl_write_sincos(glsp, &token);
 	} else {
 		struct shader_var *var = sp_getparam(glsp, token);
 		if (var && astrcmp_n(var->type, "texture", 7) == 0)
@@ -407,13 +485,13 @@ static bool gl_write_intrinsic(struct gl_shader_parser *glsp,
 }
 
 static void gl_write_function_contents(struct gl_shader_parser *glsp,
-		struct cf_token **p_token, const char *end)
+				       struct cf_token **p_token,
+				       const char *end)
 {
 	struct cf_token *token = *p_token;
 
-	if (token->type != CFTOKEN_NAME
-	    || (  !gl_write_type_token(glsp, token)
-	       && !gl_write_intrinsic(glsp, &token)))
+	if (token->type != CFTOKEN_NAME || (!gl_write_type_token(glsp, token) &&
+					    !gl_write_intrinsic(glsp, &token)))
 		dstr_cat_strref(&glsp->gl_string, &token->str);
 
 	while (token->type != CFTOKEN_NONE) {
@@ -444,7 +522,7 @@ static void gl_write_function_contents(struct gl_shader_parser *glsp,
 }
 
 static void gl_write_function(struct gl_shader_parser *glsp,
-		struct shader_func *func)
+			      struct shader_func *func)
 {
 	size_t i;
 	struct cf_token *token;
@@ -460,7 +538,7 @@ static void gl_write_function(struct gl_shader_parser *glsp,
 	dstr_cat(&glsp->gl_string, "(");
 
 	for (i = 0; i < func->params.num; i++) {
-		struct shader_var *param = func->params.array+i;
+		struct shader_var *param = func->params.array + i;
 
 		if (i > 0)
 			dstr_cat(&glsp->gl_string, ", ");
@@ -478,14 +556,14 @@ static inline void gl_write_functions(struct gl_shader_parser *glsp)
 {
 	size_t i;
 	for (i = 0; i < glsp->parser.funcs.num; i++) {
-		struct shader_func *func = glsp->parser.funcs.array+i;
+		struct shader_func *func = glsp->parser.funcs.array + i;
 		gl_write_function(glsp, func);
 	}
 }
 
-static inline void gl_write_main_interface_assign(
-		struct gl_shader_parser *glsp, struct shader_var *var,
-		const char *src)
+static inline void gl_write_main_interface_assign(struct gl_shader_parser *glsp,
+						  struct shader_var *var,
+						  const char *src)
 {
 	/* vertex shaders: write gl_Position */
 	if (glsp->type == GS_SHADER_VERTEX &&
@@ -498,12 +576,13 @@ static inline void gl_write_main_interface_assign(
 }
 
 static void gl_write_main_storage_assign(struct gl_shader_parser *glsp,
-		struct shader_var *var, const char *dst, const char *src,
-		bool input)
+					 struct shader_var *var,
+					 const char *dst, const char *src,
+					 bool input)
 {
 	struct shader_struct *st;
 	struct dstr dst_copy = {0};
-	char ch_left  = input ? '.' : '_';
+	char ch_left = input ? '.' : '_';
 	char ch_right = input ? '_' : '.';
 
 	if (dst) {
@@ -526,20 +605,31 @@ static void gl_write_main_storage_assign(struct gl_shader_parser *glsp,
 		dstr_cat_ch(&src_copy, ch_right);
 
 		for (i = 0; i < st->vars.num; i++) {
-			struct shader_var *st_var = st->vars.array+i;
+			struct shader_var *st_var = st->vars.array + i;
 			gl_write_main_storage_assign(glsp, st_var,
-					dst_copy.array, src_copy.array, input);
+						     dst_copy.array,
+						     src_copy.array, input);
 		}
 
 		dstr_free(&src_copy);
 	} else {
-		if (!dstr_is_empty(&dst_copy))
-			dstr_cat_dstr(&glsp->gl_string, &dst_copy);
-		dstr_cat(&glsp->gl_string, " = ");
-		if (src)
-			dstr_cat(&glsp->gl_string, src);
-		dstr_cat(&glsp->gl_string, var->name);
-		dstr_cat(&glsp->gl_string, ";\n");
+		if (input || (glsp->type != GS_SHADER_VERTEX) ||
+		    (strcmp(var->mapping, "POSITION"))) {
+			if (!dstr_is_empty(&dst_copy))
+				dstr_cat_dstr(&glsp->gl_string, &dst_copy);
+			dstr_cat(&glsp->gl_string, " = ");
+			if (input && (strcmp(var->mapping, "VERTEXID") == 0))
+				dstr_cat(&glsp->gl_string, "uint(gl_VertexID)");
+			else if (input && (glsp->type == GS_SHADER_PIXEL) &&
+				 (strcmp(var->mapping, "POSITION") == 0))
+				dstr_cat(&glsp->gl_string, "gl_FragCoord");
+			else {
+				if (src)
+					dstr_cat(&glsp->gl_string, src);
+				dstr_cat(&glsp->gl_string, var->name);
+			}
+			dstr_cat(&glsp->gl_string, ";\n");
+		}
 
 		if (!input)
 			gl_write_main_interface_assign(glsp, var, src);
@@ -549,28 +639,28 @@ static void gl_write_main_storage_assign(struct gl_shader_parser *glsp,
 }
 
 static inline void gl_write_main_storage_inputs(struct gl_shader_parser *glsp,
-		struct shader_func *main)
+						struct shader_func *main)
 {
 	gl_write_main_storage_assign(glsp, main->params.array, NULL,
-			"inputval_", true);
+				     "inputval_", true);
 }
 
 static inline void gl_write_main_storage_outputs(struct gl_shader_parser *glsp,
-		struct shader_func *main)
+						 struct shader_func *main)
 {
 	/* we only do this *if* we're writing a struct, because otherwise
 	 * the call to 'main' already does the assignment for us */
 	if (!main->mapping) {
 		struct shader_var var = {0};
 		var.name = "outputval";
-		var.type = (char*)main->return_type;
+		var.type = main->return_type;
 		dstr_cat(&glsp->gl_string, "\n");
 		gl_write_main_storage_assign(glsp, &var, NULL, NULL, false);
 	}
 }
 
 static inline void gl_write_main_vars(struct gl_shader_parser *glsp,
-		struct shader_func *main_func)
+				      struct shader_func *main_func)
 {
 	size_t i;
 	for (i = 0; i < main_func->params.num; i++) {
@@ -589,7 +679,7 @@ static inline void gl_write_main_vars(struct gl_shader_parser *glsp,
 }
 
 static inline void gl_write_main_func_call(struct gl_shader_parser *glsp,
-		struct shader_func *main_func)
+					   struct shader_func *main_func)
 {
 	size_t i;
 	dstr_cat(&glsp->gl_string, "\n\toutputval = _main_wrap(");
@@ -604,7 +694,7 @@ static inline void gl_write_main_func_call(struct gl_shader_parser *glsp,
 }
 
 static void gl_write_main(struct gl_shader_parser *glsp,
-		struct shader_func *main)
+			  struct shader_func *main)
 {
 	dstr_cat(&glsp->gl_string, "void main(void)\n{\n");
 
@@ -622,22 +712,22 @@ static void gl_rename_attributes(struct gl_shader_parser *glsp)
 	size_t i = 0, input_idx = 0, output_idx = 0;
 
 	for (i = 0; i < glsp->attribs.num; i++) {
-		struct gl_parser_attrib *attrib = glsp->attribs.array+i;
+		struct gl_parser_attrib *attrib = glsp->attribs.array + i;
 		struct dstr new_name = {0};
 		const char *prefix;
 		size_t val;
 
 		if (attrib->input) {
 			prefix = glsp->input_prefix;
-			val    = input_idx++;
+			val = input_idx++;
 		} else {
 			prefix = glsp->output_prefix;
-			val    = output_idx++;
+			val = output_idx++;
 		}
 
 		dstr_printf(&new_name, "%s%u", prefix, (unsigned int)val);
 		dstr_replace(&glsp->gl_string, attrib->name.array,
-				new_name.array);
+			     new_name.array);
 		dstr_move(&attrib->name, &new_name);
 	}
 }
@@ -652,7 +742,8 @@ static bool gl_shader_buildstring(struct gl_shader_parser *glsp)
 		return false;
 	}
 
-	dstr_copy(&glsp->gl_string, "#version 150\n\n");
+	dstr_copy(&glsp->gl_string, "#version 330\n\n");
+	dstr_cat(&glsp->gl_string, "const bool obs_glsl_compile = true;\n\n");
 	gl_write_params(glsp);
 	gl_write_inputs(glsp, main_func);
 	gl_write_outputs(glsp, main_func);
@@ -665,8 +756,8 @@ static bool gl_shader_buildstring(struct gl_shader_parser *glsp)
 	return true;
 }
 
-bool gl_shader_parse(struct gl_shader_parser *glsp,
-		const char *shader_str, const char *file)
+bool gl_shader_parse(struct gl_shader_parser *glsp, const char *shader_str,
+		     const char *file)
 {
 	bool success = shader_parse(&glsp->parser, shader_str, file);
 	char *str = shader_parser_geterrors(&glsp->parser);
